package org.spagic.sap.bapi;

import java.math.BigDecimal;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import javax.annotation.PostConstruct;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.NormalizedMessage;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.servicemix.bean.support.TransformBeanSupport;
import org.apache.servicemix.jbi.jaxp.SourceTransformer;
import org.apache.servicemix.jbi.jaxp.StringSource;
import org.dom4j.Document;
import org.dom4j.DocumentHelper;
import org.dom4j.xpath.DefaultXPath;

import com.sap.conn.jco.AbapException;
import com.sap.conn.jco.JCoDestination;
import com.sap.conn.jco.JCoDestinationManager;
import com.sap.conn.jco.JCoField;
import com.sap.conn.jco.JCoFunction;
import com.sap.conn.jco.JCoMetaData;
import com.sap.conn.jco.JCoParameterList;
import com.sap.conn.jco.JCoStructure;
import com.sap.conn.jco.JCoTable;

public class BAPICaller extends TransformBeanSupport {
	
	private static final Log log = LogFactory.getLog(BAPICaller.class);
	
	public String sapDestinationName = null;
	public String bapiName = null;
	private Map<String, String> namespaceMap = null;
	
	private SimpleDateFormat sapJCODateFormat = new SimpleDateFormat("yyyyMMdd"); 
	private SimpleDateFormat sapJCOTimeFormat = new SimpleDateFormat("HHmmss");
	
	//
	// The map contain as key the name of the parameter and as value the xpath rule
	// We don't need the sap type because we get it automatically
	//
	private HashMap bapiParameters  = null;
	
	private JCoDestination destination = null; 
	
	@PostConstruct
	public void init() {
		  log.info(" Initializing SAP BAPI Caller");
		  log.info(" Initializing Destination ["+sapDestinationName+"]");
		  try{
			  destination = JCoDestinationManager.getDestination(sapDestinationName);
		  }catch (Throwable e) {
			log.error("Exception in Getting Destination ", e);
		  }
	      log.info(" Initialization OK ");
	}
	
	@Override
	protected boolean transform(MessageExchange exchange, NormalizedMessage in,
			NormalizedMessage out) throws Exception {
		
		synchronized (this) {
				log.debug(" transform -> start");
				String xmlInputMessage = new SourceTransformer().contentToString(in);
				log.debug(" transform -> Input Message ");
				log.debug(" transform -> " + xmlInputMessage);
				if (destination == null){
					 log.error(" Destination is NULL Throwing Runtime Exception");
					 throw new RuntimeException("Error in Getting Destination ["+ sapDestinationName+"]");
				}
				log.debug(" transform -> Getting JCOFunction ["+bapiName+"]");
				JCoFunction function = destination.getRepository().getFunction(bapiName);
			    if(function == null){
			    	log.error(" Function ["+bapiName+"] Not Found ON SAP");	
			    	throw new RuntimeException("Function ["+ bapiName+"] Not Found on SAP");
			    }
			    log.debug(" transform -> JCOFunction is ok");
			    JCoParameterList inputParameterList = function.getImportParameterList();
			    JCoMetaData inputParametersMetadata = null;
			    try{
			    	if (inputParameterList != null){
			    		inputParametersMetadata = function.getImportParameterList().getMetaData();
			    		log.debug(" transform -> JCOFunction  Input Parameters Meta Data Retrieved " );
			    		log.debug("transform -> Set the input parameters value on jcoFunction");
				    	Set<String> keySet = bapiParameters.keySet();
				    	for (String aBapiParameter : keySet){
				    		String stringValue = evalXPath(xmlInputMessage, (String)bapiParameters.get(aBapiParameter));
				    		
				    		setValueOnFunction(aBapiParameter, function, inputParametersMetadata, stringValue);
				    	}
			    	}
			    	function.execute(destination);
			    	log.debug(" transform -> JCOFunction Execution OK");
			    	String output = jcoFunctionAsXML(function);
			    	log.debug(output);
			    	out.setContent(new StringSource(output));
			    }catch (AbapException abapException) {
					log.error("Exception calling "+function.getName(), abapException);
					throw new RuntimeException(" Abab Exception ["+abapException.getMessage()+"] during execution ["+function.getName()+"] ");
				}
		}
		
		return true;
	
	}
	 
	protected String evalXPath(String inputMessage, String xpath){
		try{
			Document document = DocumentHelper.parseText(inputMessage);
			org.dom4j.XPath xPath = new DefaultXPath(xpath);
			xPath.setNamespaceURIs(getNamespaceMap());
			return  xPath.valueOf(document);
		}catch (Exception e) {
			log.error(e);
			return null;
		}
	}
	 public  void jcoFieldAsXML(JCoField jcoField, StringBuffer sb){
	    	String jcoFieldName = jcoField.getName();
	    	int jcoFieldType = jcoField.getType();
	    	
	    	//log.debug("jcoField As XML -> Field ["+jcoFieldName+"] Type ["+jcoField.getTypeAsString()+"]"); 
	    	sb.append("<"+jcoFieldName+">");;
	    	
	    	if (jcoFieldType == JCoMetaData.TYPE_STRUCTURE){
	    		sb.append("\n");
	    		JCoStructure structure = jcoField.getStructure();
	    		for (JCoField aJCoField : structure){
	    			jcoFieldAsXML(aJCoField,sb);
	    		}
	    	}else if (jcoFieldType == JCoMetaData.TYPE_TABLE){
	    		sb.append("\n");
	    		JCoTable table = jcoField.getTable();
	    		for (int j=0; j < table.getNumRows(); j++){
	    			table.setRow(j);
	    			sb.append("<ROW idx=\""+j+"\">\n");
	    			for (JCoField aJCoField : table){
	    				jcoFieldAsXML(aJCoField,sb);
	    			}
	    			sb.append("</ROW>\n");
	    		}
	    	}else{
	    		log.debug("Field ["+jcoFieldName+"] Type ["+jcoField.getTypeAsString()+"] Value ["+jcoField.getString()+"]");
	    		sb.append(jcoField.getString());
	    	}
	    	sb.append("</"+jcoFieldName+">\n");
	    	
	    	
	    }
	    
	    public  String jcoFunctionAsXML(JCoFunction jcoFunction){
	    	
	    	JCoParameterList inputParameterList = jcoFunction.getImportParameterList();
	    	JCoParameterList resultParameterList = jcoFunction.getExportParameterList();
	    	JCoParameterList tableParameterList = jcoFunction.getTableParameterList();
	    	StringBuffer buffer = new StringBuffer();
	    	buffer.append("<"+jcoFunction.getName() + ">");
	    	// -- INPUT
	    	buffer.append("<INPUT>\n");
	    	if (inputParameterList != null){
	    		for (JCoField fld : inputParameterList){
	    			jcoFieldAsXML(fld,buffer);
	    		}
	    	}
	    	buffer.append("</INPUT>\n");
	    	
	    	// -- OUTPUT
	    	buffer.append("<OUTPUT>\n");
	    	if (resultParameterList != null){
	    		for (JCoField fld : resultParameterList){
	    			jcoFieldAsXML(fld,buffer);
	    		}
	    	}
	    	buffer.append("</OUTPUT>\n");
	    	
	    	// -- TABLES
	    	buffer.append("<TABLES>\n");
	    	if ( tableParameterList != null ){
	    		for (JCoField fld : tableParameterList){
	    			jcoFieldAsXML(fld,buffer);
	    		}
	    	}
	    	buffer.append("</TABLES>\n");
	    	buffer.append("</"+jcoFunction.getName() + ">");
	    	return buffer.toString();
	    }

		public Map<String, String> getNamespaceMap() {
			return namespaceMap;
		}

		public void setNamespaceMap(Map<String, String> namespaceMap) {
			this.namespaceMap = namespaceMap;
		}

		public String getSapDestinationName() {
			return sapDestinationName;
		}

		public void setSapDestinationName(String sapDestinationName) {
			this.sapDestinationName = sapDestinationName;
		}

		public String getBapiName() {
			return bapiName;
		}

		public void setBapiName(String bapiName) {
			this.bapiName = bapiName;
		}

		public void setValueOnFunction(String bapiParameterName, JCoFunction jcoFunction, JCoMetaData inputParametersMetadata, String stringValue) throws Exception {
			
			String className = inputParametersMetadata.getClassNameOfField(bapiParameterName);
			log.debug(" setValueOnFunction -> Name ["+bapiParameterName+"] String Value ["+stringValue+"] JCOClassName ["+className+"]");
			if (className.equalsIgnoreCase(String.class.getName())){
				
				jcoFunction.getImportParameterList().setValue(bapiParameterName, stringValue);
				log.debug(" setValueOnFunction -> setValueOnFunction as String OK");
			}else if (className.equalsIgnoreCase(BigDecimal.class.getName())){
				log.debug(" setValueOnFunction -> setValueOnFunction as BigDecimal");
				BigDecimal bdValue = new BigDecimal(stringValue);
				jcoFunction.getImportParameterList().setValue(bapiParameterName, bdValue);
				log.debug(" setValueOnFunction -> setValueOnFunction as BigDecimal OK");
			}else if (className.equalsIgnoreCase(Date.class.getName())){
				Date dateValue = null;
				if (inputParametersMetadata.getType(bapiParameterName) == JCoMetaData.TYPE_DATE){
					dateValue = sapJCODateFormat.parse(stringValue);
				}else if (inputParametersMetadata.getType(bapiParameterName) == JCoMetaData.TYPE_TIME){
					dateValue = sapJCOTimeFormat.parse(stringValue);
				}
				jcoFunction.getImportParameterList().setValue(bapiParameterName, dateValue);
				log.debug(" setValueOnFunction -> setValueOnFunction as Date");
			}else if (className.equalsIgnoreCase(Integer.class.getName())){
				int intValue = Integer.valueOf(stringValue).intValue();
				jcoFunction.getImportParameterList().setValue(bapiParameterName, intValue);
				log.debug(" setValueOnFunction -> setValueOnFunction as Int");
			}else{
				byte[] byteArrValue = stringValue.getBytes();
				jcoFunction.getImportParameterList().setValue(bapiParameterName, byteArrValue);
				log.debug(" setValueOnFunction -> setValueOnFunction as Byte Array OK");
			}
		}

		public HashMap getBapiParameters() {
			return bapiParameters;
		}

		public void setBapiParameters(HashMap bapiParameters) {
			this.bapiParameters = bapiParameters;
		}
		
		
}
